use std::collections::{HashMap, HashSet};
use std::error::Error;
use std::{io};
use std::fs::File;
use std::net::{TcpStream};
use std::path::{PathBuf};
use std::io::{BufReader, BufWriter, ErrorKind, Read, Write};
use kuchiki::iter::NodeIterator;
use serde_json::to_writer;
use crate::config::Config;

#[cfg(not(target_os = "wasi"))]
use native_tls::{TlsStream};
#[cfg(not(target_os = "wasi"))]
use imap::{Session};

#[cfg(target_os = "wasi")]
use crate::imap::imap_client::{Client, connect_to_imap_server};

mod imap_client;

#[cfg(not(target_os = "wasi"))]
pub async fn download_email_from_imap(maildir_path: PathBuf, imap_domain: String, username: String, password: String) -> Result<Vec<(u32, PathBuf)>, Box<dyn Error>>{
    let tls = native_tls::TlsConnector::builder().build().unwrap();
    let client = imap::connect((imap_domain.clone(), Config::global().imap_port.clone() as u16), imap_domain.clone(), &tls).unwrap();
    let mut imap_session = client
        .login(username, password)
        .map_err(|e| e.0)?;
    let mut stored_paths: Vec<(u32, PathBuf)> = vec![];

    match list_imap_folders(&mut imap_session).await {
        Ok(folders) => {
            for folder in folders {
                println!("Start downloading {}", folder.clone());
                let new_paths = match fetch_and_store_emails(&mut imap_session, maildir_path.clone(), folder.clone()).await {
                    Ok(paths) => {
                        println!("Emails fetched and stored successfully for {}.", folder);
                        paths
                    },
                    Err(err) => {
                        eprintln!("Error during email fetching and storing: {}", err);
                        vec![]
                    }
                };
                stored_paths.extend(new_paths);
            }
        },
        Err(e) => eprintln!("Error listing folders: {}", e),
    }

    // Logout from the IMAP session
    match imap_session.logout() {
        Ok(_) => println!("Logged out successfully."),
        Err(err) => eprintln!("Error logging out: {}", err),
    }

    Ok(stored_paths)
}

#[cfg(target_os = "wasi")]
pub async fn download_email_from_imap(maildir_path: PathBuf, imap_domain: String, username: String, password: String) -> Result<Vec<(u32, PathBuf)>, Box<dyn Error>>{
    let mut client = connect_to_imap_server(imap_domain.clone(), Config::global().imap_port.clone() as u16).await?;
    client.login(username.clone(), password.clone()).await?;
    let mut stored_paths: Vec<(u32, PathBuf)> = vec![];

    match client.list().await {
        Ok(folders) => {
            for folder in folders {
                println!("Start downloading {}", folder.clone());
                let new_paths = match fetch_and_store_emails(&mut client, maildir_path.clone(), folder.clone()).await {
                    Ok(paths) => {
                        println!("Emails fetched and stored successfully for {}.", folder);
                        paths
                    },
                    Err(err) => {
                        eprintln!("Error during email fetching and storing: {}", err);
                        vec![]
                    }
                };
                stored_paths.extend(new_paths);
            }
        },
        Err(e) => eprintln!("Error listing folders: {}", e),
    }
    
    // Logout from the IMAP session
    client.logout().await.expect("Unable to logout");

    Ok(stored_paths)
}

#[cfg(not(target_os = "wasi"))]
pub async fn list_imap_folders(imap_session: &mut Session<TlsStream<TcpStream>>) -> imap::error::Result<Vec<String>> {
    // The function return all available folders(lists) on the IMAP server
    // Fetch the list of mailboxes (folders)
    let mailboxes = imap_session.list(None, Some("*"))?;
    let folder_names = mailboxes.iter().map(|mailbox| mailbox.name().to_string()).collect();

    Ok(folder_names)
}


#[cfg(not(target_os = "wasi"))]
pub async fn fetch_and_store_emails(imap_session: &mut Session<TlsStream<TcpStream>>, maildir_path: PathBuf, list: String) -> imap::error::Result<Vec<(u32, PathBuf)>> {
    // the function download all emails from the IMAP in the folder(list)
    imap_session.select(list.clone())?;

    // creating files for maildir
    std::fs::create_dir_all(maildir_path.clone().join(list.clone()).join("new")).expect(&*("Unable to create 'new' directory in ".to_owned() + &*list.clone()));
    std::fs::create_dir_all(maildir_path.clone().join(list.clone()).join("cur")).expect(&*("Unable to create 'new' directory in ".to_owned() + &*list.clone()));
    std::fs::create_dir_all(maildir_path.clone().join(list.clone()).join("tmp")).expect(&*("Unable to create 'new' directory in ".to_owned() + &*list.clone()));

    let uids_path = maildir_path.clone().join(".uids.json");

    let uids_server = imap_session.uid_search("ALL").unwrap();
    let uids_local = read_local_uids(uids_path.clone()).unwrap();
    let mut stored_paths: Vec<(u32, PathBuf)> = vec![];
    
    for uid in uids_server {
        if let Some(uids) = uids_local.get(&list) {
            if uids.contains(&uid.to_string()) {
                continue;
            }
        }

        let messages = imap_session.uid_fetch(uid.to_string(), "(UID BODY[])")?;
        let message = messages.iter().next().ok_or(imap::error::Error::Bad("No messages found".to_string()))?;

        if let Some(body) = message.body() {
            let mail_file = store(maildir_path.clone().join(list.clone()), uid.clone().to_string(), "new".to_string(), body, "");
            match mail_file {
                Ok(file) => {stored_paths.push((uid.to_string().parse().unwrap(), file))},
                Err(e) => eprintln!("Failed to store email: {}", e),
            }
        } else {
            eprintln!("Failed to retrieve email body");
        }

        if let Some(uid) = message.uid {
            update_local_uids(uid.to_string(), list.clone(), uids_path.clone()).expect("Cannot write uid to .uids.json");
        }

        println!("Downloaded message with UID {}", uid.to_string());
    }

    Ok(stored_paths)
}

#[cfg(target_os = "wasi")]
pub async fn fetch_and_store_emails(client: &mut Client, maildir_path: PathBuf, list: String) -> Result<Vec<(u32, PathBuf)>, Box<dyn std::error::Error>> {
    // the function download all emails from the IMAP in the folder(list)
    client.select(list.clone()).await?;
    
    // creating files for maildir
    std::fs::create_dir_all(maildir_path.clone().join(list.clone()).join("new")).expect(&*("Unable to create 'new' directory in ".to_owned() + &*list.clone()));
    std::fs::create_dir_all(maildir_path.clone().join(list.clone()).join("cur")).expect(&*("Unable to create 'new' directory in ".to_owned() + &*list.clone()));
    std::fs::create_dir_all(maildir_path.clone().join(list.clone()).join("tmp")).expect(&*("Unable to create 'new' directory in ".to_owned() + &*list.clone()));
    
    let uids_path = maildir_path.clone().join(".uids.json");
    
    let uids_server = client.uid_search("ALL".to_string()).await.unwrap();
    let uids_local = read_local_uids(uids_path.clone()).unwrap();
    let mut stored_paths: Vec<(u32, PathBuf)> = vec![];
    
    for uid in uids_server {
        if let Some(uids) = uids_local.get(&list) {
            if uids.contains(&uid.to_string()) {
                continue;
            }
        }
    
        let message = client.uid_fetch(uid.clone().to_string(), "(UID BODY[])".to_string()).await?;
        let mut body = message.clone();
        
        if let Some(index) = message.find('\n') {
            body = (&message[index+1..]).to_string();
        } else {}


        let mail_file = store(maildir_path.clone().join(list.clone()), uid.clone().to_string(), "new".to_string(), body.as_bytes(), "");
        match mail_file {
            Ok(file) => {stored_paths.push((uid.to_string().parse().unwrap(), file))},
            Err(e) => eprintln!("Failed to store email: {}", e),
        }

        update_local_uids(uid.to_string(), list.clone(), uids_path.clone()).expect("Cannot write uid to .uids.txt");
        
    
        println!("Downloaded message with UID {}", uid.to_string());
    }

    Ok(stored_paths)
}


fn read_local_uids(uids_path: PathBuf) -> io::Result<HashMap<String, HashSet<String>>> {
    let file = match File::open(&uids_path) {
        Ok(file) => file,
        Err(_) => return Ok(HashMap::new()), // Propagate other errors
    };

    let mut reader = BufReader::new(file);
    let mut content = String::new();
    reader.read_to_string(&mut content)?;

    if content.trim().is_empty() {
        // Return an empty HashMap if the file is empty
        return Ok(HashMap::new());
    }

    // Deserialize the JSON into a temporary structure
    let data: HashMap<String, Vec<String>> = match serde_json::from_str(&content) {
        Ok(data) => data,
        Err(e) => return Err(io::Error::new(io::ErrorKind::InvalidData, e)), // Handle JSON parsing errors
    };

    // Convert Vec<String> to HashSet<String>
    let mut result: HashMap<String, HashSet<String>> = HashMap::new();
    for (key, value) in data {
        result.insert(key, value.into_iter().collect());
    }

    Ok(result)
}

fn update_local_uids(uid: String, list: String, uids_path: PathBuf) -> io::Result<()> {
    let file = std::fs::OpenOptions::new()
        .write(true)
        .create(true)
        .open(uids_path.clone())?;

    let mut data = read_local_uids(uids_path.clone())?;

    data.entry(list)
        .or_insert_with(HashSet::new)
        .insert(uid);

    let writer = BufWriter::new(file);

    // Convert HashSet<String> back to Vec<String> for serialization
    let mut temp_data: HashMap<String, Vec<String>> = HashMap::new();
    for (key, value) in data {
        temp_data.insert(key.clone(), value.iter().cloned().collect());
    }

    // Serialize the data back to JSON and write to file
    to_writer(writer, &temp_data)?;
    
    Ok(())
}

#[cfg(target_os = "wasi")]
pub async fn delete_email_by_uid(list: String, uid: u32) -> anyhow::Result<()>{
    let mut client = connect_to_imap_server(Config::global().imap_domain.clone(), Config::global().imap_port.clone() as u16).await?;
    client.login(Config::global().username.clone(), Config::global().password.clone()).await?;
    client.select(list.clone()).await?;
    client.mark_deleted(uid).await?;
    client.expunge().await?;
    client.logout().await?;
    Ok(())
}

#[cfg(not(target_os = "wasi"))]
pub async fn delete_email_by_uid(list: String, uid: u32) -> anyhow::Result<()>{
    let tls = native_tls::TlsConnector::builder().build().unwrap();
    let client = imap::connect((Config::global().imap_domain.clone(), Config::global().imap_port.clone() as u16), Config::global().imap_domain.clone(), &tls).unwrap();
    let mut imap_session = client
        .login(Config::global().username.clone(), Config::global().password.clone())
        .map_err(|e| e.0)?;
    
    imap_session.select(list)?;
    imap_session.uid_store(uid.to_string(), "+FLAGS (\\Deleted)")?;
    imap_session.expunge()?;
    imap_session.logout()?;

    Ok(())
}

#[cfg(not(target_os = "wasi"))]
pub async fn create_folder(name: String) -> anyhow::Result<()>{
    let tls = native_tls::TlsConnector::builder().build().unwrap();
    let client = imap::connect((Config::global().imap_domain.clone(), Config::global().imap_port.clone() as u16), Config::global().imap_domain.clone(), &tls).unwrap();
    let mut imap_session = client
        .login(Config::global().username.clone(), Config::global().password.clone())
        .map_err(|e| e.0)?;

    imap_session.create(name)?;
    imap_session.logout()?;

    Ok(())
}

#[cfg(target_os = "wasi")]
pub async fn create_folder(name: String) -> anyhow::Result<()>{
    let mut client = connect_to_imap_server(Config::global().imap_domain.clone(), Config::global().imap_port.clone() as u16).await?;
    client.login(Config::global().username.clone(), Config::global().password.clone()).await?;
    client.create_folder(name.clone()).await?;
    client.logout().await?;
    Ok(())
}

#[cfg(not(target_os = "wasi"))]
pub async fn rename_folder(name: String, new_name: String) -> anyhow::Result<()>{
    let tls = native_tls::TlsConnector::builder().build().unwrap();
    let client = imap::connect((Config::global().imap_domain.clone(), Config::global().imap_port.clone() as u16), Config::global().imap_domain.clone(), &tls).unwrap();
    let mut imap_session = client
        .login(Config::global().username.clone(), Config::global().password.clone())
        .map_err(|e| e.0)?;

    imap_session.rename(name, new_name)?;
    imap_session.logout()?;

    Ok(())
}

#[cfg(target_os = "wasi")]
pub async fn rename_folder(name: String, new_name: String) -> anyhow::Result<()>{
    let mut client = connect_to_imap_server(Config::global().imap_domain.clone(), Config::global().imap_port.clone() as u16).await?;
    client.login(Config::global().username.clone(), Config::global().password.clone()).await?;
    client.rename_folder(name.clone(), new_name.clone()).await?;
    client.logout().await?;
    Ok(())
}

#[cfg(not(target_os = "wasi"))]
pub async fn delete_folder(name: String) -> anyhow::Result<()>{
    let tls = native_tls::TlsConnector::builder().build().unwrap();
    let client = imap::connect((Config::global().imap_domain.clone(), Config::global().imap_port.clone() as u16), Config::global().imap_domain.clone(), &tls).unwrap();
    let mut imap_session = client
        .login(Config::global().username.clone(), Config::global().password.clone())
        .map_err(|e| e.0)?;

    imap_session.delete(name)?;
    imap_session.logout()?;

    Ok(())
}

#[cfg(target_os = "wasi")]
pub async fn delete_folder(name: String) -> anyhow::Result<()>{
    let mut client = connect_to_imap_server(Config::global().imap_domain.clone(), Config::global().imap_port.clone() as u16).await?;
    client.login(Config::global().username.clone(), Config::global().password.clone()).await?;
    client.delete_folder(name.clone()).await?;
    client.logout().await?;
    Ok(())
}

fn store(
    path: PathBuf,
    uid: String,
    subfolder: String,
    data: &[u8],
    info: &str,
) -> io::Result<PathBuf> {
    // loop when conflicting filenames occur, as described at
    // http://www.courier-mta.org/maildir.html
    // this assumes that pid and hostname don't change.
    let mut tmppath = path.clone();
    tmppath.push("tmp");

    let mut file;
    loop {

        tmppath.push(format!("email_{uid}"));

        match std::fs::OpenOptions::new()
            .write(true)
            .create_new(true)
            .open(&tmppath)
        {
            Ok(f) => {
                file = f;
                break;
            }
            Err(err) => {
                if err.kind() != ErrorKind::AlreadyExists {
                    return Err(err.into());
                }
                tmppath.pop();
            }
        }
    }

    /// At this point, `file` is our new file at `tmppath`.
    /// If we leave the scope of this function prior to
    /// successfully writing the file to its final location,
    /// we need to ensure that we remove the temporary file.
    /// This struct takes care of that detail.
    struct UnlinkOnError {
        path_to_unlink: Option<PathBuf>,
    }

    impl Drop for UnlinkOnError {
        fn drop(&mut self) {
            if let Some(path) = self.path_to_unlink.take() {
                // Best effort to remove it
                std::fs::remove_file(path).ok();
            }
        }
    }

    // Ensure that we remove the temporary file on failure
    let mut unlink_guard = UnlinkOnError {
        path_to_unlink: Some(tmppath.clone()),
    };

    file.write_all(data)?;
    file.sync_all()?;

    let mut newpath = path.clone();
    newpath.push(subfolder);


    let id = format!("email_{uid}");
    newpath.push(format!("{}{}", id, info));

    std::fs::rename(&tmppath, &newpath)?;
    unlink_guard.path_to_unlink.take();
    Ok(newpath)
}