use mail_parser::{Message};
use std::collections::{HashMap, HashSet};
use std::error::Error;
use std::path::{PathBuf};
use models::*;
use std::ffi::{OsStr, OsString};
use std::fs;
use std::fs::{create_dir_all, File, OpenOptions};
use std::io::{BufReader, Write};
use std::net::SocketAddr;
use anyhow::anyhow;
use serde::Deserialize;
use crate::indexes::{Indexes, SerializableMessage, SerializableThread};
use config::{Config, INSTANCE};
use crate::server::{get_email, get_folders, get_sorted_threads_by_date, get_thread_messages};
use mailparse::parse_mail;
use regex::Regex;
use crate::templates::util::parse_email;

#[cfg(not(target_os = "wasi"))]
use axum::extract::Query;
#[cfg(not(target_os = "wasi"))]
use axum::{Json, Router};
#[cfg(not(target_os = "wasi"))]
use axum::http::Method;
#[cfg(not(target_os = "wasi"))]
use axum::response::{IntoResponse, Response};
#[cfg(not(target_os = "wasi"))]
use axum::routing::{get, post};
#[cfg(not(target_os = "wasi"))]
use tower_http::cors::{Any, CorsLayer};

#[cfg(target_os = "wasi")]
use bytecodec::DecodeExt;
#[cfg(target_os = "wasi")]
use httpcodec::{HttpVersion, ReasonPhrase, Request, RequestDecoder, Response, StatusCode, Header};
#[cfg(target_os = "wasi")]
use lettre::message::{Mailbox, Message as LettreMessage, MultiPart, SinglePart};
#[cfg(target_os = "wasi")]
use wasmedge_wasi_socket::{Shutdown, TcpListener, TcpStream};
#[cfg(target_os = "wasi")]
use crate::server_wasi::{handle_client};
#[cfg(target_os = "wasi")]
use crate::smtp_client::{send_email};

mod smtp_client;

mod arg;
mod config;
mod models;
mod templates;
mod util;
mod indexes;
mod imap;
mod server;
mod server_wasi;

pub fn append_ext(ext: impl AsRef<OsStr>, path: &PathBuf) -> PathBuf {
    let mut os_string: OsString = path.into();
    os_string.push(".");
    os_string.push(ext.as_ref());
    os_string.into()
}

fn write_if_changed<T: AsRef<[u8]>>(path: &PathBuf, data: T) -> bool {
    if let Ok(d) = std::fs::read(path) {
        if d == data.as_ref() {
            return false;
        }
    }

    std::fs::write(path, data).unwrap();
    true
}

pub fn message_to_html(msg: StrMessage) -> String{
    fn remove_style_tags(html: &str) -> String {
        Regex::new(r"(?s)<style.*?>.*?</style>").unwrap()
            .replace_all(html, "")
            .to_string()
    }
    fn extract_body(html: &str) -> String {
        let open_tag = "<body";
        let close_tag = "</body>";

        if let Some(start_pos) = html.find(open_tag) {
            if let Some(open_tag_end) = html[start_pos..].find('>') {
                let body_start = start_pos + open_tag_end + 1;
                if let Some(end_pos) = html[body_start..].find(close_tag) {
                    return html[body_start..body_start + end_pos].to_string();
                }
            }
        }
        html.to_string()
    }
    fn extract_styles(html: &str) -> String {
        let style_pattern = Regex::new(r"(?is)<style[^>]*>(.*?)</style>").unwrap();
        style_pattern.captures_iter(html)
            .filter_map(|caps| caps.get(1))
            .map(|m| m.as_str())
            .collect::<Vec<_>>()
            .join("\n")
    }


    let data = match fs::read_to_string(&msg.original_path.clone()) {
        Ok(content) => content,
        Err(e) => {
            eprintln!("Error reading file: {}", e);
            return "".to_string();
        }
    };

    let parsed_mail = match parse_mail(&data.as_bytes()) {
        Ok(mail) => mail,
        Err(e) => return "".to_string(),
    };

    let html = parse_email(&parsed_mail).unwrap_or_else(|_| msg.body.clone());
    let styles = extract_styles(&html);
    let mut body_content = remove_style_tags(&extract_body(&html));
    // body_content = purple_numbers(&*body_content, "#");

    format!(
        r#"
                <!DOCTYPE html>
                <html>
                <head>
                    <title>Email</title>
                    <meta http-equiv='Permissions-Policy' content='interest-cohort=()'/>
                    <meta name='viewport' content='width=device-width, initial-scale=1.0, maximum-scale=1.0,user-scalable=0' />
                </head>      
                <body>         
                    <div id="email-body" class="email-body" style="position: relative;">
                        <style>
                            {styles}
                        </style>
                        <div">
                            {body_content}
                        </div>
                    </div>
                </body>
                </html>
            "#,
        styles = styles,
        body_content = body_content,
    )
}

fn persist_email(msg: &Message, uid: u32, list: String, original_path: PathBuf) -> anyhow::Result<()>{
    let mut indexes = Indexes::new();
    
    // creating all folders
    if Config::global().out_dir.exists() {
        std::fs::create_dir_all(&Config::global().out_dir).ok();
    }
    let message_dir = Config::global().out_dir.join(list.clone()).join("messages");
    let message_xml_dir = Config::global().out_dir.join(list.clone()).join("messages_xml");
    if !message_dir.exists() {
        std::fs::create_dir_all(&message_dir).ok();
    }
    if !message_xml_dir.exists() {
        std::fs::create_dir_all(&message_xml_dir).ok();
    }

    // persist email
    let mut message = StrMessage::new(msg, uid, list.clone());
    message.set_original_path(original_path);
    let mut hash_path = message_dir.clone().join(message.hash.clone());
    hash_path = append_ext("html", &hash_path.clone());
    let relative_path = match hash_path.clone().strip_prefix(Config::global().out_dir.clone()) {
        Ok(relative_path) => {PathBuf::from(relative_path)}
        Err(_) => {hash_path.clone()}
    };
    message.set_path(relative_path.display().to_string());
    if Config::global().include_html {
        let html = append_ext("html", &message_dir.join(message.hash.clone()));
        let mut file = File::create(html)?;
        let data = message_to_html(message.clone());
        file.write_all(data.as_bytes())?;
        file.flush()?;
    }

    // xml
    let xml = append_ext("xml", &message_xml_dir.join(message.hash.clone()));
    let mut file = File::create(xml)?;
    let data = message.to_xml();
    file.write_all(data.as_bytes())?;
    file.flush()?;
    
    // indexing
    let _ = Indexes::add(message.clone());
    
    Ok(())
}

fn parse_args(){
    let args = arg::Args::from_env();
    
    let mut config = Config::from_file(&args.config).expect("could not read config");
    if !args.include_html { config.include_html = true; } 
    else { config.include_html = args.include_html; }
    
    config.maildir = args.maildir.clone();
    config.out_dir = args.out_dir.clone();
    INSTANCE.set(config).unwrap();
}

async fn run(){
    parse_args();

    // downloading new emails
    let new_paths = imap::download_email_from_imap(
        Config::global().maildir.clone(),
        Config::global().imap_domain.clone(),
        Config::global().username.clone(),
        Config::global().password.clone()
    ).await.expect("Cannot download new emails");

    
    let mut lists: Vec<String> = Vec::new();
    
    // reading the folder with lists
    for mail_list in std::fs::read_dir(Config::global().maildir.clone())
        .expect("could not read maildir")
        .filter_map(|m| m.ok())
    {
        let dir_name = mail_list.file_name().into_string().unwrap();
        
        if dir_name.starts_with('.') || ["cur", "new", "tmp"].contains(&dir_name.as_str()) {
            continue;
        }
        lists.push(dir_name.clone());
        
        if !Config::global().out_dir.clone().join(dir_name.clone()).exists(){
            std::fs::create_dir_all(Config::global().out_dir.clone().join(dir_name.clone())).ok();
        }
        
        // let mut uid = 1; // fake uid for tests
        // for entry in std::fs::read_dir(Config::global().maildir.clone().join(mail_list.path().clone()).join("new")).expect("could not read maildir").filter_map(|m| m.ok()){
        for (uid, path) in new_paths.clone().into_iter().filter(|(u16, path)| path.starts_with(mail_list.path())) {
            // let path = entry.path();
            
            let data = std::fs::read(&path).expect("could not read mail");
            
            if let Some(mail) = Message::parse(&data) {
                let _ = persist_email(&mail, uid.clone(), dir_name.clone(), path.clone());
            } else {
                println!("Could not parse message {:?}", path);
            }
            // uid += 1;
        }
    }

    let threads_indexes_path = Config::global().out_dir.clone().join("threads.json");
    if !threads_indexes_path.exists() {
        Indexes::persist_threads().expect("Unable to persist threads");
        Indexes::persist_indexes(lists).expect("Unable to persist indexes");    
    }
}

async fn delete_email(message_id: String) -> anyhow::Result<()>{
    let message = Indexes::delete_from_messages(message_id);
    match message {
        Ok(message) => {
            let _ = Indexes::delete_from_threads(message.clone());
            let lists: Vec<String> = fs::read_dir(Config::global().out_dir.clone())?
                .filter_map(Result::ok)
                .filter(|entry| entry.metadata().map(|m| m.is_dir()).unwrap_or(false))
                .filter_map(|entry| entry.file_name().into_string().ok())
                .collect();
            let _ = Indexes::persist_indexes(lists);

            let message_html_name = Config::global().out_dir.clone()
                .join(message.list.clone())
                .join("messages")
                .join(message.hash.clone() + ".html");
            let message_xml_name = Config::global().out_dir.clone()
                .join(message.list.clone())
                .join("messages_xml")
                .join(message.hash.clone() + ".xml");
            let _ = fs::remove_file(message_html_name);
            let _ = fs::remove_file(message_xml_name);
                
            let _ = imap::delete_email_by_uid(message.list.clone(), message.uid.clone()).await;
        }
        Err(_) => {}
    }
    
    Ok(())
}

#[cfg(target_os = "wasi")]
#[tokio::main(flavor = "current_thread")]
async fn main() -> anyhow::Result<()> {
    run().await;


    // let email = LettreMessage::builder()
    //     .from("sokolovskiiyura@outlook.com".parse().unwrap())
    //     .to("sokolovskiiyura@gmail.com".parse().unwrap())
    //     .subject("Hello, World!")
    //     .multipart(
    //         MultiPart::alternative() // This is composed of two parts.
    //             .singlepart(
    //                 SinglePart::plain("Hello, World!".to_string()) // Plain text version
    //             )
    //             .singlepart(
    //                 SinglePart::html("<h1>Hello, World!</h1>".to_string()) // HTML version
    //             )
    //     )
    //     .unwrap();
    
    
    // let _ = send_email(email).await;
    
    
    // delete_email("Sent".to_string(), 4).await;
    
    // imap::create_folder("testFolder".to_string()).await.unwrap();
    // imap::rename_folder("testFolder".to_string(), "newTestFolder".to_string()).await.unwrap();
    // imap::delete_folder("newTestFolder".to_string()).await.unwrap();
    
    
    // API
    println!("Server is running on {}", Config::global().api_port);
    let listener = TcpListener::bind(format!("{}:{}", Config::global().api_addr, Config::global().api_port), false)?;
    loop {
        let (stream, _) = listener.accept(false)?;
        let _ = handle_client(stream).await;
    }

    Ok(())
}

#[cfg(not(target_os = "wasi"))]
#[tokio::main]
async fn main() {
    run().await;
    
    // let email = LettreMessage::builder()
    //     .from("sokolovskiiyura@outlook.com".parse().unwrap())
    //     .to("sokolovskiiyura@gmail.com".parse().unwrap())
    //     .subject("Hello, World!")
    //     .multipart(
    //         MultiPart::alternative() // This is composed of two parts.
    //             .singlepart(
    //                 SinglePart::plain("Hello, World!".to_string()) // Plain text version
    //             )
    //             .singlepart(
    //                 SinglePart::html("<h1>Hello, World!</h1>".to_string()) // HTML version
    //             )
    //     )
    //     .unwrap();
 
    
    // let _ = send_email(email).await;

    
    // delete_email("lqo7m8r2.1js7w080jvr2o@express.medallia.com".to_string()).await.expect("Unable to delete an email");


    // imap::create_folder("testFolder".to_string()).await.unwrap();
    // imap::rename_folder("testFolder".to_string(), "newTestFolder".to_string()).await.unwrap();
    // imap::delete_folder("newTestFolder".to_string()).await.unwrap();
    
    
    // API
    // Define the CORS layer
    let cors = CorsLayer::new()
        .allow_origin(Any)
        .allow_methods(vec![Method::GET, Method::POST])
        .allow_headers(Any);
    
    let app = Router::new()
        .route("/folders", get(get_folders_handle))
        .route("/sorted_threads_by_date", get(sorted_threads_by_date_handle))
        .route("/get_thread_messages", get(get_thread_messages_handle))
        .route("/email", get(get_email_handle))
        .layer(cors);
    
    let listener = tokio::net::TcpListener::bind(format!("{}:{}", Config::global().api_addr, Config::global().api_port)).await.unwrap();
    println!("Server is running on {}", Config::global().api_port);
    axum::serve(listener, app).await.unwrap();
}

#[cfg(not(target_os = "wasi"))]
#[derive(Deserialize)]
struct GetSortedThreadsByDateQuery {
    folder: String,
}
#[cfg(not(target_os = "wasi"))]
async fn sorted_threads_by_date_handle(Query(params): Query<GetSortedThreadsByDateQuery>) -> Json<Vec<(String, Vec<u32>)>> {
    let result: Vec<(String, Vec<u32>)> = serde_json::from_str(&*get_sorted_threads_by_date(params.folder).await).unwrap();
    Json(result)
}

#[cfg(not(target_os = "wasi"))]
async fn get_folders_handle() -> Json<Vec<String>> {
    let result: Vec<String>  = serde_json::from_str(&*get_folders().await).unwrap();
    Json(result)
}

#[cfg(not(target_os = "wasi"))]
#[derive(Deserialize)]
struct GetEmailQuery {
    id: String,
}
#[cfg(not(target_os = "wasi"))]
async fn get_email_handle(Query(params): Query<GetEmailQuery>) -> Response {
    let result: String  = get_email(params.id).await;
    result.into_response()
}

#[cfg(not(target_os = "wasi"))]
#[derive(Deserialize)]
struct GetThreadMessagesQuery {
    id: u32,
}
#[cfg(not(target_os = "wasi"))]
async fn get_thread_messages_handle(Query(params): Query<GetThreadMessagesQuery>) -> Json<Vec<SerializableMessage>> {
    let result: Vec<SerializableMessage> = serde_json::from_str(&*get_thread_messages(params.id).await).unwrap();
    Json(result)
}
