#![cfg(target_os = "wasi")]

use std::collections::HashMap;
use std::io::{Read, Write};
use anyhow::anyhow;
use serde::de::Unexpected::Str;
use bytecodec::DecodeExt;
use httpcodec::{HttpVersion, ReasonPhrase, Request, RequestDecoder, Response, StatusCode, Header};
use wasmedge_wasi_socket::{Shutdown, TcpListener, TcpStream};
use crate::server::{get_email, get_folders, get_sorted_threads_by_date, get_thread_messages};

async fn handle_http(req: Request<String>, request_line: &str) -> anyhow::Result<String> {
    let parts: Vec<&str> = request_line.split_whitespace().collect();
    if parts.len() < 3 {
        return Err(anyhow!("Invalid Input"))
    }
    let (path, query) = parse_url(parts[1]);

    // Basic CORS headers
    let cors_headers = "HTTP/1.1 200 OK\r\n\
    Content-Type: application/json\r\n\
    Access-Control-Allow-Origin: *\r\n\
    Access-Control-Allow-Methods: GET, POST, PUT, DELETE, OPTIONS\r\n\
    Access-Control-Allow-Headers: Content-Type, Authorization\r\n";

    if req.method().to_string().eq("OPTIONS") {
        return Ok(format!("{}Content-Length: {}\r\n", cors_headers, 0))
    }

    
    let content = match path.as_str() {
        "/folders" => format!("{}", get_folders().await),
        "/sorted_threads_by_date" => {
            if let Some(value) = query.get("folder"){
                format!("{}", get_sorted_threads_by_date(value.to_owned()).await)
            }else{ "404 Not Found".to_string() }
        },
        "/get_thread_messages" => {
            if let Some(value) = query.get("id"){
                match value.parse::<u32>() {
                    Ok(value) => format!("{}", get_thread_messages(value).await),
                    Err(_e) => "404 Not Found".to_string(),
                }
            }else{"404 Not Found".to_string()}
        }
        "/email" => {
            if let Some(value) = query.get("id"){
                format!("{}", get_email(value.to_owned()).await)
            }else{ "404 Not Found".to_string() }
        },
        _ => "404 Not Found".to_string(),
    };

    Ok(format!("{}Content-Length: {}\r\n\r\n{}", cors_headers, content.len(), content))
}

fn parse_url(url: &str) -> (String, HashMap<String, String>) {
    // Split the string into path and query parts
    let parts: Vec<&str> = url.splitn(2, '?').collect();

    let path = parts[0].to_string();
    let mut query_params = HashMap::new();

    if parts.len() > 1 {
        // Split the query part into key-value pairs
        let query_part = parts[1];
        for param in query_part.split('&') {
            let kv: Vec<&str> = param.splitn(2, '=').collect();
            if kv.len() == 2 {
                query_params.insert(kv[0].to_string(), kv[1].to_string());
            } else {
                query_params.insert(kv[0].to_string(), String::new());
            }
        }
    }

    (path, query_params)
}

pub async fn handle_client(mut stream: TcpStream) -> std::io::Result<()> {
    let mut buffer = [0u8; 1024];
    let mut data = Vec::new();

    while let Ok(n) = stream.read(&mut buffer) {
        if n == 0 {
            break; // Connection was closed
        }
        data.extend_from_slice(&buffer[..n]);
        if n < buffer.len() { break; } // Less data means end of request
    }

    // Assuming the first line is the request line
    let request_line = std::str::from_utf8(&data)
        .map_err(|e| std::io::Error::new(std::io::ErrorKind::InvalidData, e))?
        .lines()
        .next()
        .ok_or(std::io::Error::new(std::io::ErrorKind::InvalidData, "No request line"))?;

    let mut decoder = RequestDecoder::<httpcodec::BodyDecoder<bytecodec::bytes::Utf8Decoder>>::default();
    let req = decoder.decode_from_bytes(&data)
        .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, e))?;

    let response = handle_http(req, request_line)
        .await
        .map_err(|e| std::io::Error::new(std::io::ErrorKind::Other, e))?;

    let write_buf = response.to_string();
    stream.write_all(write_buf.as_bytes())?;
    stream.shutdown(Shutdown::Both)?;
    Ok(())
}